module SPC_Tests (tests) where

import Control.Concurrent (threadDelay)
import Data.IORef
import SPC
import Test.Tasty (TestTree, localOption, mkTimeout, testGroup)
import Test.Tasty.HUnit (testCase, (@?=))
import Data.Either (isRight)

tests :: TestTree
tests =
  localOption (mkTimeout 3000000) $
    testGroup
      "SPC (core)"
      [
        testCase "workerAdd" $ do
          spc <- startSPC

          w <- workerAdd spc "R2-D2"
          isRight w @?= True,
        testCase "workerAdd (2)" $ do
          spc <- startSPC

          w1 <- workerAdd spc "MSE-6"
          isRight w1 @?= True

          w2 <- workerAdd spc "GNK"
          isRight w2 @?= True,
        testCase "workerAdd (fail)" $ do
          spc <- startSPC

          w1 <- workerAdd spc "BD-1"
          isRight w1 @?= True

          w2 <- workerAdd spc "BD-1"
          isRight w2 @?= False,
        testCase "Running a job" $ do
          ref <- newIORef False
          spc <- startSPC

          w <- workerAdd spc "R5-D4"
          isRight w @?= True

          j <- jobAdd spc $ Job (writeIORef ref True) 1
          r <- jobWait spc j
          r @?= Done

          x <- readIORef ref
          x @?= True,
        testCase "Running two jobs" $ do
          ref <- newIORef (0::Int)
          spc <- startSPC

          w <- workerAdd spc "K-2SO"
          isRight w @?= True

          j1 <- jobAdd spc $ Job (writeIORef ref 1) 1
          r1 <- jobWait spc j1
          r1 @?= Done

          x1 <- readIORef ref
          x1 @?= 1

          j2 <- jobAdd spc $ Job (writeIORef ref 2) 1
          r2 <- jobWait spc j2
          r2 @?= Done

          x2 <- readIORef ref
          x2 @?= 2,
        testCase "Canceling job (pending)" $ do
          spc <- startSPC
          j <- jobAdd spc $ Job (pure ()) 1
          jobCancel spc j
          r <- jobStatus spc j
          r @?= JobDone DoneCancelled,
        testCase "Canceling job (running)" $ do
          spc <- startSPC

          w <- workerAdd spc "IG-88"
          isRight w @?= True

          j <- jobAdd spc $ Job (threadDelay 2000000) 2
          jobCancel spc j
          r <- jobStatus spc j
          r @?= JobDone DoneCancelled,
        testCase "Canceling job (running) (new job)" $ do
          ref <- newIORef False
          spc <- startSPC

          w <- workerAdd spc "C-3PO"
          isRight w @?= True

          j1 <- jobAdd spc $ Job (threadDelay 2000000) 2
          jobCancel spc j1
          r1 <- jobStatus spc j1
          r1 @?= JobDone DoneCancelled

          -- job has been cancelled. Starting new job
          j2 <- jobAdd spc $ Job (writeIORef ref True) 1
          r2 <- jobWait spc j2
          r2 @?= Done

          x <- readIORef ref
          x @?= True,
        testCase "Timeout" $ do
          spc <- startSPC

          w <- workerAdd spc "L3-37"
          isRight w @?= True

          j <- jobAdd spc $ Job (threadDelay 2000000) 1
          r <- jobWait spc j

          r @?= DoneTimeout,
        testCase "Timeout (2 jobs)" $ do
          ref <- newIORef False
          spc <- startSPC

          w <- workerAdd spc "General Kalani"
          isRight w @?= True

          j1 <- jobAdd spc $ Job (threadDelay 2000000) 1
          j2 <- jobAdd spc $ Job (writeIORef ref True) 1
          r1 <- jobWait spc j1
          r1 @?= DoneTimeout

          r2 <- jobWait spc j2
          r2 @?= Done

          x <- readIORef ref
          x @?= True,
        testCase "Crash" $ do
          ref <- newIORef False
          spc <- startSPC
          w <- workerAdd spc "C1-10P"
          isRight w @?= True

          j1 <- jobAdd spc $ Job (error "boom") 1
          r1 <- jobWait spc j1
          r1 @?= DoneCrashed

          -- Ensure new jobs can still work.
          j2 <- jobAdd spc $ Job (writeIORef ref True) 1
          r2 <- jobWait spc j2
          r2 @?= Done
          x <- readIORef ref
          x @?= True,
        testCase "Remove worker" $ do
          spc <- startSPC
          w1 <- workerAdd spc "D-O"
          isRight w1 @?= True
          case w1 of
            (Right worker) -> do
              w2 <- workerAdd spc "D-O" -- Can't make another worker with same name yet
              isRight w2 @?= False

              j <- jobAdd spc $ Job (threadDelay 1000000) 1
              workerStop worker
              threadDelay 100
              r <- jobStatus spc j
              r @?= JobDone DoneCancelled

              w3 <- workerAdd spc "D-O" -- But we can make one now
              isRight w3 @?= True
            _ -> False @?= True
      ]
